import pandas as pd
import numpy as np
import math

def create_city_zone_coordinates(path_to_json_data):

    """ Calls the functions import_geojson and calc_centroids to get
    a list of city cone coordinates and created a pandas DataFrame in the 
    required format, i.e. wiht columns zone_lat and zone_long, as well as
    zone_id as index column
    """

    # map Uber Movement zone IDs to lists of lats and longs
    (
        map_movement_id_to_latitude_coordinates,
        map_movement_id_to_longitude_coordinates
    ) = import_geojson(path_to_json_data)

    # calculate centroids of city zone polygons
    (
        map_movement_id_to_centroid_lat,
        map_movement_id_to_centroid_long
    ) = calc_centroids(
       map_movement_id_to_latitude_coordinates,
        map_movement_id_to_longitude_coordinates
    )

    # create a pandas Dataframe in required format for bevpo
    city_zone_coordinates = pd.DataFrame(
        list(map_movement_id_to_centroid_lat.items()), 
        columns = ['zone_id','zone_lat']
    )
    
    city_zone_coordinates['zone_long'] = (
        map_movement_id_to_centroid_long.values()
    )
    
    # set the zone_id column as index
    city_zone_coordinates.set_index('zone_id', inplace=True)

    return city_zone_coordinates


def import_geojson(path_to_json_data):

    """ Imports the geojson data from the passed path and maps Uber Movement
    city zone IDs to a flattened list of latitude and longitude coordinates
    in the format of two dictionaries. Uses the recursive function called
    foster_coordinates_recursive to flatten the differently nested data.
    """
    
    data = pd.read_json(path_to_json_data)
    data.pop('type')
    data = data['features']
    
    map_json_entry_to_movement_id = dict()

    for json_id, json_entry in enumerate(data):
        
        map_json_entry_to_movement_id[json_id] = int(
          json_entry['properties']['MOVEMENT_ID']
        )
    
    map_movement_id_to_latitude_coordinates = dict()
    map_movement_id_to_longitude_coordinates = dict()

    for k, v in map_json_entry_to_movement_id.items():
        map_movement_id_to_latitude_coordinates[v] = []
        map_movement_id_to_longitude_coordinates[v] = []


    for json_id, movement_id in map_json_entry_to_movement_id.items():
        coordinates = data[json_id]['geometry']['coordinates']
        
        (
            map_movement_id_to_latitude_coordinates, 
            map_movement_id_to_longitude_coordinates
        ) = foster_coordinates_recursive(
            movement_id,
            map_movement_id_to_latitude_coordinates,
            map_movement_id_to_longitude_coordinates,
            coordinates
        )
        
    
    map_movement_id_to_coordinates = (
        map_movement_id_to_latitude_coordinates,
        map_movement_id_to_longitude_coordinates
    )

    return map_movement_id_to_coordinates


def foster_coordinates_recursive(
    movement_id,
    map_movement_id_to_latitude_coordinates,
    map_movement_id_to_longitude_coordinates,
    coordinates
):

    """ Flattens the coordinates of a passed city zone id (movement_id)
    and coordiates list recursively and saves their numeric values
    in the dictionaries that map movement ids to a list of latitude and 
    longitude coordinates.
    """

    dummy = 0

    for j in coordinates:

        if type(j) != list and dummy == 0:

            map_movement_id_to_longitude_coordinates[movement_id].append(j)
            dummy = 1
            continue

        elif type(j) != list and dummy == 1:

            map_movement_id_to_latitude_coordinates[movement_id].append(j)
            break

        else:

            dummy = 0
            coordinates = j
            (
                map_movement_id_to_latitude_coordinates,
                map_movement_id_to_longitude_coordinates
            ) = foster_coordinates_recursive(
                movement_id,
                map_movement_id_to_latitude_coordinates,
                map_movement_id_to_longitude_coordinates,
                coordinates
            )

    map_movement_id_to_coordinates = (
        map_movement_id_to_latitude_coordinates,
        map_movement_id_to_longitude_coordinates
    )

    return map_movement_id_to_coordinates


def calc_centroids(
    map_movement_id_to_latitude_coordinates,
    map_movement_id_to_longitude_coordinates
):

    """ Calculates the centroid of passed city zone polygons. Should a city
    zone consist of unregularities or multiple polygons, this is identified
    by centroid coordinates that are not within the bound of minimum and 
    maximum values of all coordinates of that city zone. In this case, the
    centroids are replaced with the mean of lat and long coordinates.
    """
    
    # create empty dictionary for mapping Uber Movement IDs to city zone areas
    map_movement_id_to_cityzone_area = dict()

    # iterate over all movement IDs and latitude coordinates
    for movement_id, lat_coordinates in map_movement_id_to_latitude_coordinates.items():
        
        # get also the longitude coordinates
        long_coordinates = map_movement_id_to_longitude_coordinates[movement_id]
        
        # calculate currently iterated city zone area
        area_cityzone = 0
        for i in range(len(lat_coordinates)-1):

            area_cityzone = (
                area_cityzone
                + long_coordinates[i] * lat_coordinates[i+1]
                - long_coordinates[i+1] * lat_coordinates[i]
            )
      
        area_cityzone = (
            area_cityzone
            + long_coordinates[i+1] * lat_coordinates[0]
            - long_coordinates[0] * lat_coordinates[i+1]
        )
        
        area_cityzone *= 0.5
        #area_cityzone = abs(area_cityzone)
        
        map_movement_id_to_cityzone_area[movement_id] = area_cityzone
        
    # create empty dictionaries for mapping Uber Movement IDs to city zone centroids
    map_movement_id_to_centroid_lat = dict()
    map_movement_id_to_centroid_long = dict()
        
    # iterate over all movement IDs and latitude coordinates
    for movement_id, lat_coordinates in map_movement_id_to_latitude_coordinates.items():
        
        # get also the longitude coordinates
        long_coordinates = map_movement_id_to_longitude_coordinates[movement_id]
        
        
        # calculate currently iterated city zone area
        centroid_lat = 0
        centroid_long = 0
        for i in range(len(lat_coordinates)-1):
            
            centroid_long += (
                long_coordinates[i]
                + long_coordinates[i+1]
            ) * (
                long_coordinates[i] * lat_coordinates[i+1]
                - long_coordinates[i+1] * lat_coordinates[i]
            )

            centroid_lat += (
                lat_coordinates[i]
                + lat_coordinates[i+1]
            ) * (
                long_coordinates[i] * lat_coordinates[i+1]
                - long_coordinates[i+1] * lat_coordinates[i]
            )

        centroid_long += (
            long_coordinates[i+1]
            + long_coordinates[0]
        ) * (
            long_coordinates[i+1] * lat_coordinates[0]
            - long_coordinates[0] * lat_coordinates[i+1]
        )
        
        centroid_lat += (
                lat_coordinates[i+1]
                + lat_coordinates[0]
            ) * (
                long_coordinates[i+1] * lat_coordinates[0]
                - long_coordinates[0] * lat_coordinates[i+1]
            )
        

        centroid_lat /= (
            6 * map_movement_id_to_cityzone_area[movement_id]
        )
        centroid_long /= (
            6 * map_movement_id_to_cityzone_area[movement_id]
        )
     
        # Uber Movement city zones sometimes consist of multiple distinct polygons
        if (
            centroid_lat < min(lat_coordinates)
            or centroid_lat > max(lat_coordinates)
            or centroid_long < min(long_coordinates)
            or centroid_long > max(long_coordinates)
        ):
            # in this case we calculate the mean instead of centroid
            centroid_lat = np.mean(lat_coordinates)
            centroid_long = np.mean(long_coordinates)            
        
        map_movement_id_to_centroid_lat[movement_id] = centroid_lat
        map_movement_id_to_centroid_long[movement_id] = centroid_long
        
    map_movement_id_to_centroid_coordinates = (
        map_movement_id_to_centroid_lat,
        map_movement_id_to_centroid_long
    )
    
    return map_movement_id_to_centroid_coordinates
    

def create_od_matrix_lists(path_to_rawdata):

    """ imports the raw Uber Movement travel time data and creates tw0 lists of
    origin destination (OD) matrices. A first list contains the mean travel time
    and a distinct second list contains the standard deviation of travel time. 
    Each matrix and therefore each list element corresponds to a single time 
    step. 
    """
    
    rename_dict_mean = {
        'sourceid': 'source_id',
        'dstid': 'dest_id'
    }
    
    rename_dict_stddev = {
        'sourceid': 'source_id',
        'dstid': 'dest_id',
        'standard_deviation_travel_time': 'stddev_travel_time'
    }
    
    travel_data = pd.read_csv(path_to_rawdata)
    od_mean_travel_time_list = []
    od_std_travel_time_list = []

    T = 24
    for t in range(T):
        hod_travel_data = travel_data.loc[travel_data['hod'] == t]
        od_mean_travel_time_list.append(
            hod_travel_data[
                ['sourceid', 'dstid','mean_travel_time']
            ].rename(columns=rename_dict_mean)
        )
        od_std_travel_time_list.append(
            hod_travel_data[
                ['sourceid', 'dstid','standard_deviation_travel_time']
            ].rename(columns=rename_dict_stddev)
        )
    
    
    od_travel_time_lists = (
        od_mean_travel_time_list,
        od_std_travel_time_list
    )
    
    return od_travel_time_lists



import bevpo.prob_dist as prob_dist
import bevpo.samp_traf as samp_traf
import bevpo.calc_tfsprop as calc_tfsprop
import bevpo.save_results as save_results

import math
import pandas as pd
import numpy as np

class TrafficSystem:

    """ Bundles all properties and methods of the traffic system
    modeled with bevpo.
    """

    def __init__(
        self,
        city_zone_coordinates,
        od_mean_travel_time_list,
        od_stddev_travel_time_list=None,
        od_distances=None,
        charging_profile=None,
        e_drive=2,
        e_dest=2,
        p_min=0.1,
        p_max=0.9,
        cars_per_zone=10
    ):

        ### Parameters
        self.city_zone_coordinates = city_zone_coordinates
        self.od_mean_travel_time_list = od_mean_travel_time_list
        self.od_stddev_travel_time_list = od_stddev_travel_time_list
        self.od_distances = od_distances
        self.charging_profile = charging_profile
        self.e_drive = e_drive
        self.e_dest = e_dest
        self.p_min = p_min
        self.p_max = p_max
        self.cars_per_zone = cars_per_zone
        
        ### Attributes
        self.T = len(od_mean_travel_time_list)
        self.number_zones = len(city_zone_coordinates)
        self.datatensor_mean = 0
        self.datatensor_stddev = 0
        self.C = round(
            self.number_zones * self.cars_per_zone
        )
        self.state_tensor = np.zeros(
            (
                self.C,
                self.T
            )
        ).astype(int)
        self.transition_tensor = np.zeros(
            (
                self.C,
                self.T,
                4 # driving x destination x travel time x distance
            )
        )
        # if no origin-destination travel distances passed,
        # calculate beeline distance between city zone centroids
        if od_distances is None:
            self.calc_od_distances(city_zone_coordinates)
        
        ### Results placeholders
        self.driving_map = 0
        self.parking_map = 0
        self.charging_map = 0
        self.avg_driving_times = 0
        self.avg_driving_distances = 0
        self.driving_share_lifetime = 0
        self.parking_share_lifetime = 0
        self.circadian_rhythm = 0
        self.distr_distances_per_t = 0
        self.distr_durations_per_t = 0
        self.distr_distances_total = 0
        self.distr_durations_total = 0
        self.distr_bins_km = 0
        self.distr_bins_s = 0
        self.charging_profile_dist = 0
        
        
    def calc_od_distances(self, city_zone_coordinates):
    
        """ Calculates the beeline distance between origin destination zones
        of a city if no explicit matrix of such distances was past when
        initializing class object.
        """
        
        c_lat_long = 111.3
        conv_deg_rad = 0.01745

        zone_id_array= city_zone_coordinates.index.values
        od_distances = pd.DataFrame(0, index=zone_id_array, columns=zone_id_array)

        # iterate over all city zones
        for entry_num1, zone_id1 in enumerate(zone_id_array):
            
            lat1 = city_zone_coordinates.loc[zone_id1]['zone_lat']
            long1 = city_zone_coordinates.loc[zone_id1]['zone_long']
            
            # iterate over all zones up to currently iterated one
            for entry_num2 in range(entry_num1+1):
                
                zone_id2 = zone_id_array[entry_num2]
                lat2 = city_zone_coordinates.loc[zone_id2]['zone_lat']
                long2 = city_zone_coordinates.loc[zone_id2]['zone_long']
                
                # calculate beeline distance
                distance_km = (
                    c_lat_long * math.sqrt(
                        (
                            math.cos(
                                (
                                    lat1
                                    + lat2
                                ) / 2 * conv_deg_rad
                            )
                        )**2 * (
                            long1
                            - long2
                        )**2
                        + (
                            lat1
                            - lat2
                        )**2
                    )
                )
                
                # write matrix entries symmetrically with 10% additional 
                # distance compared to the beeline
                od_distances.loc[zone_id1, zone_id2] = distance_km * 1.1
                od_distances.loc[zone_id2, zone_id1] = distance_km * 1.1
            
            # overwrite the diagonal entries with 1 km distances
            od_distances.loc[zone_id1, zone_id2] = 1
            od_distances.loc[zone_id2, zone_id1] = 1
            
        self.od_distances = od_distances
        

    def simulate_traffic(self):

        """ Simulates the traffic system when called. """
        
        ### Transform data from list of dataframes into single datatensor
        self.create_datatensors()
        
        ### Calculate distributions of driving and choosind a destination
        prob_dist.calc_prob_dists(self)
        
        ### Sample traffic
        samp_traf.sample_traffic(self)
        
        ### Calculate traffic system properties
        calc_tfsprop.calc_traffic_system_properties(self)
        

    def create_datatensors(self):

        """ Transforms the list of od matrices into a single datatensor. 
        City zone indexing corresponds to respective index positions in
        city_zone_coordinates.index. Time step indexing corresponds to 
        respective position of OD matrix in od_mean_travel_time_list.
        If od_stddev_travel_time_list is available, the same is done for this 
        list too.
        """

        datatensor_mean = np.zeros(
            (
                self.number_zones,
                self.number_zones,
                self.T
            )
        )
        # iterate over all time steps
        for time in range(self.T):
            # get od matrix of current time step
            od_mean_travel_time_df = self.od_mean_travel_time_list[time]

            # iterate over all rows in od matric
            for index, row in od_mean_travel_time_df.iterrows():
                # get the mean travel time column only
                mean = row['mean_travel_time']
                # set the source id to position where zone coordinate index matches
                source = int(
                    np.where(
                        self.city_zone_coordinates.index.values == row['source_id']
                    )[0]
                )
                # set also destination id to position where zone coordinate 
                # index matches
                dest = int(
                    np.where(
                        self.city_zone_coordinates.index.values == row['dest_id']
                    )[0]
                )
                # assign mean value to respective datatensor entry
                datatensor_mean[source, dest, time] = mean

        # set the list of od matrices to zero for saving memory
        self.od_mean_travel_time_list = 0
        # save extracted datatensor as class object attribute
        self.datatensor_mean = datatensor_mean

        # do the same for od standard deviations of travel time if available
        if self.od_stddev_travel_time_list is not None:
            datatensor_stddev = np.zeros(
                (
                    self.number_zones,
                    self.number_zones,
                    self.T
                )
            )

            for time in range(self.T):
                od_stddev_travel_time_df = self.od_stddev_travel_time_list[time]

                for index, row in od_stddev_travel_time_df.iterrows():
                    stddev = row['stddev_travel_time']
                    source = int(
                        np.where(
                            self.city_zone_coordinates.index.values == row['source_id']
                        )[0]
                    )
                    dest = int(
                        np.where(
                            self.city_zone_coordinates.index.values == row['dest_id']
                        )[0]
                    )

                datatensor_stddev[source, dest, time] = stddev

            self.od_stddev_travel_time_list = 0
            self.datatensor_stddev = datatensor_stddev
            
            
    def save_tfs_results(self, path_to_results=None):
    
        """ Saves the resulting plots and numeric values unter path_to_results
        when called.
        """
        
        # set path to current folder if no path is declared
        if path_to_results is None:
            path_to_results = './bevpo_results/'
            
        # set path as attribute of class object
        self.path_to_results = path_to_results
        
        # call function from bevpo.save_results.py module
        save_results.save_results(self)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import imageio
import os
from mpl_toolkits import mplot3d

def save_results(tfs):

    """ Creates results folders, csv and plots and saves these in the respective
    folders.
    """
    
    create_resultsfolder(tfs)
    save_as_csv(tfs)
    save_as_plots(tfs)
    
    save_maps('parking', tfs, tfs.parking_map, 150000, 'parking_map.gif')
    save_maps('driving', tfs, tfs.driving_map, 150000, 'driving_map.gif')
    if tfs.charging_profile is not None:
        save_maps('charging', tfs, tfs.charging_map, 150000, 'charging_map.gif')


def create_resultsfolder(tfs):

    """ Creates a folder called bevpo_results on the tfs.path_to_results for
    saving both the plots and numeric values of the traffic simulation outcome.
    """

    if tfs.path_to_results.endswith('/'):
        tfs.path_to_bevpo_results = tfs.path_to_results
    else:
        tfs.path_to_bevpo_results = tfs.path_to_results + '/'

    tfs.path_to_csv = tfs.path_to_bevpo_results + 'numeric/'
    tfs.path_to_figures = tfs.path_to_bevpo_results + 'figures/'
    tfs.path_to_images = tfs.path_to_bevpo_results + 'images/'
    tfs.path_to_png = tfs.path_to_images + 'png/'
    tfs.path_to_pdf = tfs.path_to_images + 'pdf/'
    
    # create required folders for saving results if not existent yet
    if not os.path.isdir(tfs.path_to_bevpo_results):
        os.mkdir(tfs.path_to_bevpo_results)
    if not os.path.isdir(tfs.path_to_csv):
        os.mkdir(tfs.path_to_csv)
    if not os.path.isdir(tfs.path_to_figures):
        os.mkdir(tfs.path_to_figures)
    if not os.path.isdir(tfs.path_to_images):
        os.mkdir(tfs.path_to_images)
    if not os.path.isdir(tfs.path_to_png):
        os.mkdir(tfs.path_to_png)
    if not os.path.isdir(tfs.path_to_pdf):
        os.mkdir(tfs.path_to_pdf)
    
                                   
def save_as_csv(tfs):

    """ Saves the numeric results to csv files."""
    
    ### Save driving and  parking maps
    df_columns = []
    for t in range(tfs.T):
        column = 't={}'.format(t)
        df_columns.append(column)
        
    saving_path = tfs.path_to_csv + 'driving_map.csv'
    df = pd.DataFrame(
        tfs.driving_map, 
        index=tfs.city_zone_coordinates.index.values,
        columns=df_columns
    )
    df.to_csv(
        saving_path,
        index_label='zone_id'
    )
    
    saving_path = tfs.path_to_csv + 'parking_map.csv'
    df = pd.DataFrame(
        tfs.parking_map,
        index=tfs.city_zone_coordinates.index.values,
        columns=df_columns
    )
    df.to_csv(
        saving_path,
        index_label='zone_id'
    )
    
    ### Save charging maps for electric vehicles
    if tfs.charging_profile is not None:
        saving_path = tfs.path_to_csv + 'charging_map.csv'
        df = pd.DataFrame(
            tfs.charging_map,
            index=tfs.city_zone_coordinates.index.values,
            columns=df_columns
        )
        df.to_csv(
            saving_path,
            index_label='zone_id'
        )
    
    ### Save traffic system properties
    saving_path = tfs.path_to_csv + 'traffic_properties.csv'
    
    df_index = []
    df_index.append('total')
    for t in range(tfs.T):
        index = 't={}'.format(t)
        df_index.append(index)
        
    df_columns=[
        'avg driving time (min)',
        'avg driving distance (km)',
        'circadian rhythm (%)'
    ]
    
    avg_properties = np.column_stack(
        (
            tfs.avg_driving_times, 
            tfs.avg_driving_distances,
            np.append(
                '',
                np.round(tfs.circadian_rhythm * 100).astype(int)
            )
        )
    )
    df = pd.DataFrame(
        avg_properties, 
        columns=df_columns,
        index=df_index
    )
    df['driving share lifetime (%)'] = ''
    df['parking share lifetime (%)'] = ''
    
    df.iloc[0, 3] = tfs.driving_share_lifetime
    df.iloc[0, 4]  = tfs.parking_share_lifetime
    
    df.to_csv(saving_path)
    
    ### Save travel distributions
    saving_path = tfs.path_to_csv + 'distance_distributions.csv'
    df_index = []
    df_index.append('total (%)')
    for t in range(tfs.T):
        index = 't={}'.format(t)
        df_index.append(index)
    
    df_columns = []
    for bin_value in tfs.distr_bins_km:
        column = '{} km'.format(round(bin_value))
        df_columns.append(column)
    
    distr_distances = np.row_stack(
        (
            tfs.distr_distances_total, 
            tfs.distr_distances_per_t
        )
    )
    df = pd.DataFrame(
        distr_distances, 
        columns=df_columns,
        index=df_index
    )
    df.to_csv(saving_path)
    
    saving_path = tfs.path_to_csv + 'duration_distributions.csv'
    df_index = []
    df_index.append('total (%)')
    for t in range(tfs.T):
        index = 't={}'.format(t)
        df_index.append(index)
    
    df_columns = []
    for bin_value in tfs.distr_bins_s:
        column = '{} s'.format(round(bin_value))
        df_columns.append(column)
    
    distr_durations = np.row_stack(
        (
            tfs.distr_durations_total, 
            tfs.distr_durations_per_t
        )
    )
    df = pd.DataFrame(
        distr_durations, 
        columns=df_columns,
        index=df_index
    )
    df.to_csv(saving_path)


def save_maps(
    title, 
    tfs, 
    tfs_map, 
    scatter_factor, 
    gif_filename,
    fig_size=13,
    font=16
):

    """ Creates driving, parking and charging maps. """
    # create folders for saving files first
    tfs.path_to_png_map = tfs.path_to_png + title + '/'
    tfs.path_to_pdf_map = tfs.path_to_pdf + title + '/'
    
    if not os.path.isdir(tfs.path_to_png_map):
        os.mkdir(tfs.path_to_png_map)
    if not os.path.isdir(tfs.path_to_pdf_map):
        os.mkdir(tfs.path_to_pdf_map) 
    
    # save png and pdf images of the map first, and keep files in filename_list
    filename_list = []
    for t in range(tfs.T):

        plt.figure(
            figsize=(fig_size, fig_size)
        )

        plt.scatter(
            tfs.city_zone_coordinates['zone_long'].values,
            tfs.city_zone_coordinates['zone_lat'].values,
            tfs_map[:, t] * scatter_factor,
            alpha=0.7
        )
        plt.title(title + f'\n t={t}')
        plt.xlabel('longitude')
        plt.ylabel('latitude')

        filename_png = f't={t}.png'
        filename_pdf = f't={t}.pdf'
        filename_list.append(filename_png)
        saving_path_png = tfs.path_to_png_map + filename_png
        saving_path_pdf = tfs.path_to_pdf_map + filename_pdf
        plt.savefig(saving_path_png)
        plt.savefig(saving_path_pdf)
        plt.close()

    # create a gif from map images
    path_to_gif = tfs.path_to_figures + gif_filename

    with imageio.get_writer(path_to_gif, mode='I', fps=3) as writer:
        for filename in filename_list:
            loading_path = tfs.path_to_png_map + filename
            image = imageio.imread(loading_path)
            writer.append_data(image)


def save_as_plots(
    tfs,
    fig_size=13,
    font=16,
):

    """ Saves the resulting statistics to .png and .pdf plots."""  

    ### Save circadian rhythm
    
    saving_path_pdf = tfs.path_to_figures + 'circadian_rhythm.pdf'
    saving_path_png = tfs.path_to_figures + 'circadian_rhythm.png'
    fig = plt.figure(
        figsize=(fig_size, fig_size)
    )
    plt.plot(
        np.round(
            tfs.circadian_rhythm * 100
        ).astype(int)
    )
    plt.title(
        'Circadian rhythm of traffic',
        fontsize=font
    )
    plt.xlabel(
        'time',
        fontsize=font
    )
    plt.ylabel(
        'traffic activity [%]',
        fontsize=font
    )
    plt.xticks(
        fontsize=font
    )
    plt.yticks(
        fontsize=font
    )
    fig.savefig(saving_path_pdf)
    fig.savefig(saving_path_png)
    plt.close(fig)
    
    ### Save provided charging profile
    if tfs.charging_profile is not None:
        saving_path_pdf = tfs.path_to_figures + 'charging_profile.pdf'
        saving_path_png = tfs.path_to_figures + 'charging_profile.png'
        fig = plt.figure(
            figsize=(fig_size, fig_size)
        )
        plt.plot(
            np.round(
                tfs.charging_profile_dist * 100
            ).astype(int)
        )
        plt.title(
            'EV charging distribution over time',
            fontsize=font
        )
        plt.xlabel(
            'time',
            fontsize=font
        )
        plt.ylabel(
            'charging [%]',
            fontsize=font
        )
        plt.xticks(
            fontsize=font
        )
        plt.yticks(
            fontsize=font
        )
        fig.savefig(saving_path_pdf)
        fig.savefig(saving_path_png)
        plt.close(fig)    
    
    
    ### Save travel distance distributions
    
    saving_path_pdf = tfs.path_to_figures + 'travel_distance.pdf'
    saving_path_png = tfs.path_to_figures + 'travel_distance.png'
    fig = plt.figure(
        figsize=(fig_size, fig_size)
    )
    plt.bar(
        range(len(tfs.distr_bins_km)),
        tfs.distr_distances_total
    )
    plt.xticks(
        range(len(tfs.distr_bins_km)),
        np.round(tfs.distr_bins_km).astype(int),
        fontsize=font
    )
    
    plt.title(
        'Travelled distances per trip',
        fontsize=font
    )
    plt.xlabel(
        'distance [km]',
        fontsize=font
    )
    plt.yticks(
        fontsize=font
    )
    plt.ylabel(
        'trips [%]',
        fontsize=font
    )
    fig.savefig(saving_path_pdf)
    fig.savefig(saving_path_png)
    plt.close(fig)
    
    
    ### Save travel duration distributions
    
    saving_path_pdf = tfs.path_to_figures + 'travel_duration.pdf'
    saving_path_png = tfs.path_to_figures + 'travel_duration.png'
    fig = plt.figure(
        figsize=(fig_size, fig_size)
    )
    plt.bar(
        range(len(tfs.distr_bins_s)),
        tfs.distr_durations_total
    )
    plt.xticks(
        range(len(tfs.distr_bins_s)),
        np.round(tfs.distr_bins_s/60, 2),
        fontsize=font
    )
    plt.yticks(
        fontsize=font
    )
    plt.title(
        'Travelled durations per trip',
         fontsize=font
    )
    plt.xlabel(
        'duration [min]',
        fontsize=font
    )
    plt.ylabel(
        'trips [%]',
        fontsize=font
    )
    fig.savefig(saving_path_pdf)
    fig.savefig(saving_path_png)
    plt.close(fig)
    
    
    saving_path_pdf = tfs.path_to_figures + 'travel_distance_per_t.pdf'
    saving_path_png = tfs.path_to_figures + 'travel_distance_per_t.png'
    fig = plt.figure(
        figsize=(fig_size, fig_size)
    )
    ax = plt.axes(projection='3d')
    ax.plot_surface(
        tfs.distr_bins_km,
        np.expand_dims(np.arange(tfs.T), 1),
        tfs.distr_distances_per_t,
    )
    ax.set_title(
        'Travel distance distribution',
        fontsize=font
    )
    ax.set_xlabel(
        'travel distance [km]',
        fontsize=font
    )
    ax.set_ylabel(
        'time step (t)',
        fontsize=font
    )
    ax.set_zlabel(
        'sampled cars',
        fontsize=font
    )
    fig.savefig(saving_path_png)
    fig.savefig(saving_path_pdf)
    plt.close(fig)
    
    saving_path_pdf = tfs.path_to_figures + 'travel_duration_per_t.pdf'
    saving_path_png = tfs.path_to_figures + 'travel_duration_per_t.png'
    fig = plt.figure(
        figsize=(fig_size, fig_size)
    )
    ax = plt.axes(projection='3d')
    ax.plot_surface(
        tfs.distr_bins_s,
        np.expand_dims(np.arange(tfs.T), 1),
        tfs.distr_durations_per_t,
    )
    ax.set_title(
        'Travel duration distribution',
        fontsize=font
    )
    ax.set_xlabel(
        'travel time [s]',
        fontsize=font
    )
    ax.set_ylabel(
        'time step (t)',
        fontsize=font
    )
    ax.set_zlabel(
        'sampled cars',
        fontsize=font
    )
    fig.savefig(saving_path_png)
    fig.savefig(saving_path_pdf)
    plt.close(fig)
        



import numpy as np

def calc_prob_dists(tfs):

    """ calculates the probabilities of driving p_drive and choosing a 
    destination p_dest.
    """


    create_distribution_p_drive(tfs)
    create_distribution_p_dest(tfs)
    #create_distribution_p_joint(tfs)


def create_distribution_p_drive(tfs):

    """ Calculates binomial probability distributions of driving for 
    each city zone and each time step. Note that city zones correspond
    to positions in datatensor and not the origional IDs from the
    city_zone_coordinates.index array.
    """

    p_drive = np.zeros(
        (
            tfs.number_zones,
            tfs.T
        )
    )
    mean_sum = np.zeros(
        (
            tfs.number_zones,
            tfs.T
        )
    )
    min_t = np.zeros(
        (
            tfs.number_zones
        )
    )
    max_t = np.zeros(
        (
            tfs.number_zones
        )
    )
    for source in range(tfs.number_zones):

        for time in range(tfs.T):
            mean_sum_term = 0
            counter = 0

            for dest in range(tfs.number_zones):
                if tfs.datatensor_mean[source, dest, time] > 0:
                    counter += 1
                    mean_sum_term = (
                        mean_sum_term
                        + tfs.datatensor_mean[source, dest, time] / (
                            tfs.od_distances.iloc[source, dest]
                        )
                    )

            if counter > 0:
                mean_sum[source, time] = mean_sum_term / counter

        value_array = mean_sum[source, :]
        min_t[source] = np.amin(value_array)
        max_t[source] = np.amax(value_array)

    for source in range(tfs.number_zones):
        if max_t[source] >0:

            for time in range(tfs.T):
                if mean_sum[source, time] > 0:
                    value = (
                        tfs.p_min
                        + (tfs.p_max - tfs.p_min) * (
                            (mean_sum[source, time] - min_t[source]) /
                            (max_t[source] - min_t[source])
                        )**tfs.e_drive
                    )
                else:
                    value = 0

                p_drive[source, time] = value

    tfs.p_drive = p_drive


def create_distribution_p_dest(tfs):

    """ Calculates probability distributions of choosing a destination.
    Note that city zones correspond to positions in datatensor and not 
    the origional IDs from the city_zone_coordinates.index array.
    """

    p_dest = np.zeros(
        (
            tfs.number_zones,
            tfs.number_zones,
            tfs.T
        )
    )
    min_x = np.zeros(
        (
            tfs.number_zones,
            tfs.number_zones
        )
    )
    max_x = np.zeros(
        (
            tfs.number_zones,
            tfs.number_zones
        )
    )
    normalization_factor = np.zeros(
        (
            tfs.number_zones,
            tfs.T
        )
    )
    for source in range(tfs.number_zones):

        for dest in range(tfs.number_zones):
            value_array = tfs.datatensor_mean[source, dest, :]
            max_x[source, dest] = np.amax(
                value_array
            )
            min_x[source, dest] = np.amin(
                value_array
            )

    for source in range(tfs.number_zones):

        for dest in range(tfs.number_zones):

            for time in range(tfs.T):
                if max_x[source, dest] > 0:
                    mean = tfs.datatensor_mean[source, dest, time]
                    p_dest[source, dest, time] = (
                        (
                            (mean - min_x[source, dest]) / (
                                max_x[source, dest] - min_x[source, dest]
                            )
                        )**tfs.e_dest
                    )

    for source in range(tfs.number_zones):

        for time in range(tfs.T):
            value_array = p_dest[source, :, time]
            normalization_factor[source, time] = np.sum(
                value_array
            )

    for source in range(tfs.number_zones):

        for time in range(tfs.T):
            if normalization_factor[source, time] > 0:
            
                for dest in range(tfs.number_zones):
                    p_dest[source, dest, time] = (
                        p_dest[source, dest, time] / (
                            normalization_factor[source, time]
                        )
                    )

    tfs.p_dest = p_dest


def create_distribution_p_joint(tfs):

    """ Calculates the joint probability distribution of driving and
    choosing a destination. Note that city zones correspond to positions 
    in datatensor and not the origional IDs from thecity_zone_coordinates.index 
    array.
    """
    p_joint = np.zeros(
        (
            tfs.number_zones,
            tfs.number_zones,
            tfs.T
        )
    )
    for source in range(tfs.number_zones):

        for time in range(tfs.T):
            multiplication_term = tfs.p_drive[source, time]
            if multiplication_term != 0:

                for dest in range(tfs.number_zones):
                    p_joint[source, dest, time] = (
                        tfs.p_dest[source, dest, time] * (
                            multiplication_term
                        )
                    )

    tfs.p_joint = p_joint
    
    



import numpy as np
import random


def sample_traffic(tfs):

    """ Initializes the traffic system state and then samples traffic for
    all time steps.
    """

    # solve initial value problem for traffic state
    solve_initial_value_problem(tfs)

    # simluate over all time steps
    for t in range(tfs.T):
        driving_activity_sampling(
            tfs,
            t
        )
        destination_choice_sampling(
            tfs,
            t
        )
        traveltime_and_distance_sampling(
            tfs,
            t
        )
        # update state matrix only up to last time step,
        # but sample transition matrix one step beyond last.
        if t < tfs.T-1:
            tfs.state_tensor[:, t+1] = tfs.transition_tensor[:, t, 1]

    # set distributions to zero for saving memory
    tfs.p_drive = 0
    tfs.p_dest = 0
    tfs.p_joint = 0

def solve_initial_value_problem(tfs):

    """ Initializes the traffic system state by first distributing all cars
    uniformly accross all city zones and then sampling through all time steps.
    The resulting system state after an entire sampling process is then taken
    as the initial system state. """

    # uniformly assign cars to to all city zones for state in initial time step
    zone = 0
    for car in range(0, tfs.C, tfs.cars_per_zone):
        tfs.state_tensor[car:(car+tfs.cars_per_zone), 0] = zone
        zone += 1
        
    # transform to integer values
    tfs.state_tensor = tfs.state_tensor.astype(int)
    
    # sample over all time steps
    for t in range(tfs.T):
        driving_activity_sampling(
            tfs,
            t
        )
        destination_choice_sampling(
            tfs,
            t
        )

        if t < tfs.T-1:
            tfs.state_tensor[:, t+1] = tfs.transition_tensor[:, t, 1]
        else:
            tfs.state_tensor[:, 0] = tfs.transition_tensor[:, -1, 1].astype(int)


def driving_activity_sampling(
    tfs,
    t
):

    """ Samples if a car drives or stays parked in a respective city zone
    from p_drive.
     """

    # iterate over all cars we simulate
    for car in range(tfs.C):
        # set origin zone to location of car in t
        origin = tfs.state_tensor[car, t]
        # get driving probability for origin-time combination
        driving_probability = tfs.p_drive[origin, t]
        # sample if car drives from this probability
        drive = np.random.choice(
            [1, 0],
            p=[driving_probability, 1-driving_probability]
        )
        # assig sampling outcome to transition matrix of car-timestep combination
        tfs.transition_tensor[car, t, 0] = drive
        # if car stays parked, also assign destination zone to stay orgin
        # in the transition matrix of car-timestep combination
        if drive == 0:
            tfs.transition_tensor[car, t, 1] = origin


def destination_choice_sampling(
    tfs,
    t
):

    """ Samples travel destinations from p_dest. """

    # iterate over all cars
    for car in range(tfs.C):
        
        # get binary sampling result from driving_activity_sampling() 
        drive = tfs.transition_tensor[car, t, 0]
        
        # only if car is sampled to drive, make a change to transition matrix
        # otherwise, we have already assigned destination to stay origin of 
        # car in in tfs.transition_tensor during driving_activity_sampling().
        if drive == 1:
            # get location of car
            origin = tfs.state_tensor[car, t]
            # get probability distribution of location-timestep combination
            distribution = tfs.p_dest[origin, :, t]
            # if distribution sum is zero, car has destination in same zone
            if sum(distribution) == 0:
                destination = origin
            else:
                # otherwise, car chooses destination with probabilities 
                # from corresponding distribution
                destination = np.random.choice(
                    range(tfs.number_zones),
                    p=distribution
                )
                
            # save the sampling outcome in transition matrix entry of
            # car-timestep combination
            tfs.transition_tensor[car, t, 1] = destination


def traveltime_and_distance_sampling(
    tfs,
    t
):

    """ Samples travel times and distances from normal Gaussian distributions.
    """

    # iterate over all cars
    for car in range(tfs.C):
        # get if car is sampled to drive
        drive = tfs.transition_tensor[car, t, 0]

        # sample driving duration and distance only if car drives
        if drive == 1:
            # get origin-destination sampling results
            origin = tfs.state_tensor[car, t]
            destination = int(
                round(
                    tfs.transition_tensor[car, t, 1]
                )
            )

            # if car drives within same city zone, use assumptions
            if origin == destination:
                # assumption: five minutes of travel time
                tfs.transition_tensor[car, t, 2] = 5 * 60
                # assumption: 1 km of travel distance
                tfs.transition_tensor[car, t, 3] = 1
            else:
                # Sampling travel times
                mean = tfs.datatensor_mean[origin, destination, t]
                if type(tfs.datatensor_stddev) != int:
                    std_deviation = tfs.datatensor_stddev[origin, destination, t]
                else:
                    std_deviation = 1 # CAUTION: assumption

                tfs.transition_tensor[car, t, 2] = abs(
                    np.random.normal(
                        mean,
                        std_deviation
                    )
                )
                
                # Sampling travel distances
                mean = tfs.od_distances.iloc[origin, destination]
                std_deviation = 0.1 # CAUTION: assumption
                tfs.transition_tensor[car, t, 3] = abs(
                    np.random.normal(
                        mean,
                        std_deviation
                    )
                )
                





import numpy as np
import matplotlib.pyplot as plt
import imageio


def calc_traffic_system_properties(tfs):

    """ Calculates a series of traffic system properties resulting from the 
    sampling. 
    """
    
    create_parking_and_driving_maps(tfs)
    calc_traffic_properties(tfs)
    calc_travel_distributions(tfs)
    calc_charging_distributions(tfs)
    

def calc_charging_distributions(tfs):

    """ Calculates the distribution of charging over time and space by splitting
    the passed charging profile over parked cars in each time step. CAUTION 
    ASSUMPTION: parked cars are charged at every given time.
    """
    
    if tfs.charging_profile is not None:
    
        if tfs.T != len(tfs.charging_profile):
        
            print(
                'Provided charging profile does not match number of time steps'
            )
            
        else:

            # turn charging profile into a numpy array
            charging_profile = np.array(tfs.charging_profile)
            
            # min-max scale provided charging_profile
            charging_profile_minmax = (
                charging_profile - min(charging_profile) / (
                    max(charging_profile) - min(charging_profile)
                )
            )
            
            # scale to valid distribution that sums to 1
            charging_profile_dist = (
                charging_profile / sum(charging_profile)
            )
            
            # copy the distribution of parking cars
            tfs.charging_map = tfs.parking_map.copy()
            
            # iterate over timesteps and scale parking cars to charging profile
            for t in range(tfs.T):
                tfs.charging_map[:, t] /= charging_profile_dist[t]
                
            # scale distribution to unit sum 
            tfs.charging_map /= np.sum(tfs.charging_map)
            
            # save the scaled charging profile
            tfs.charging_profile_dist = charging_profile_dist
    

def create_parking_and_driving_maps(tfs):

    """ Normalizes state and state transition matrices and produces
    the resulting parking and driving maps.
    """

    driving_map = np.zeros(
        (
            tfs.number_zones, 
            tfs.T
        )
    )
    parking_map = np.zeros(
        (
            tfs.number_zones, 
            tfs.T
        )
    )
    for car in range(tfs.C):
    
        for t in range(tfs.T):
            zone = tfs.state_tensor[car, t]
            driving_activity = tfs.transition_tensor[car, t, 0]

            if driving_activity == 1:
                driving_map[zone-1, t] += 1
            else:
                parking_map[zone-1, t] += 1

    tfs.driving_map = driving_map / np.sum(driving_map)
    tfs.parking_map = parking_map / np.sum(parking_map)


def calc_traffic_properties(tfs):

    """ Calculates statistics on average driving times and distances and driving 
    and parking shares per lifetime, as well as circadian rhythm of traffic. 
    """
    
    # two values for driving time, distance
    avg_properties = np.zeros(
        (
            tfs.T+1,
            2
        )
    )
    for t in range(tfs.T):
        driving_time_t = tfs.transition_tensor[:, t, 2] / 60
        driving_distance_t = tfs.transition_tensor[:, t, 3]
        avg_properties[t+1, 0] = np.mean(driving_time_t)
        avg_properties[t+1, 1] = np.mean(driving_distance_t)
        
    avg_properties[0, 0] = np.mean(avg_properties[1:, 0])
    avg_properties[0, 1] = np.mean(avg_properties[1:, 1])
    
    # calculate driving and parking shares per lifetime
    driving_share_lifetime = round(
        avg_properties[0, 0] / 60 * 100
    )
    parking_share_lifetime = 100 - driving_share_lifetime
    
    
    ### Calculuate circadian rhythm
    circadian_rhythm = np.sum(
        tfs.driving_map,
        axis=0
    )
    min_traf = np.min(circadian_rhythm)
    max_traf = np.max(circadian_rhythm)
    circadian_rhythm = (circadian_rhythm - min_traf) / (max_traf - min_traf)

    ### save results to class object attributes
    tfs.avg_driving_times = avg_properties[:, 0]
    tfs.avg_driving_distances = avg_properties[:, 1]
    tfs.driving_share_lifetime = driving_share_lifetime
    tfs.parking_share_lifetime = parking_share_lifetime
    tfs.circadian_rhythm = circadian_rhythm


def calc_travel_distributions(tfs):
    
    """ calculates the multivariate distribution of travelled distances and 
    durations for both fine-grained time steps and all time steps together.
    """
    
    # calculate overall distributions
    duration_histogram = np.histogram(
        tfs.transition_tensor[:, :, 2]
    )
    distance_histogram = np.histogram(
        tfs.transition_tensor[:, :, 3]
    )
    bins_km = distance_histogram[1][1:]
    bins_s = duration_histogram[1][1:]
    distr_durations_total = np.round(
        duration_histogram[0] / sum(duration_histogram[0]) * 100,
        2
    )
    distr_distances_total = np.round(
        distance_histogram[0] / sum(distance_histogram[0]) * 100,
        2
    )
    distr_durations_per_t = np.zeros(
        (
            tfs.T,
            len(bins_s)
        )
    )
    distr_distances_per_t = np.zeros(
        (
            tfs.T,
            len(bins_km)
        )
    )
    
    # calculate distributions over each time step
    for t in range(tfs.T):
        durations_t = tfs.transition_tensor[:, t, 2]
        distances_t = tfs.transition_tensor[:, t, 3]
        distr_durations_t = np.histogram(
            durations_t,
            len(bins_s),
            (
                bins_s[0],
                bins_s[-1]
            )
        )[0]
        distr_distances_t = np.histogram(
            distances_t,
            len(bins_km),
            (
                bins_km[0],
                bins_km[-1]
            )
        )[0]
        
        distr_durations_per_t[t, :] = distr_durations_t
        distr_distances_per_t[t, :] = distr_distances_t
        
    # assign as class properties
    tfs.distr_distances_per_t = distr_distances_per_t
    tfs.distr_durations_per_t = distr_durations_per_t
    tfs.distr_distances_total = distr_distances_total
    tfs.distr_durations_total = distr_durations_total
    tfs.distr_bins_km = bins_km
    tfs.distr_bins_s = bins_s
    
      
        
